from gotointerpreter import interpret
from unittest import mock
from unittest import TestCase


def input_continue(prompt):
    return ''


def str_yes(prompt):
    return 'J'


def input_exit(prompt):
    return 'EXIT'


class GOTOInterpreterTest(TestCase):
    def test_assignment_default_zero(self):
        self.assertEqual(0, interpret('M1:x0:=x0 + 0;M2:HALT;'))
        self.assertEqual(0, interpret('M1:x0:=x1 + 0;M2:HALT;'))
        self.assertEqual(0, interpret('M1:x0:=x2 + 0;M2:HALT;'))

    def test_assignment_non_negative(self):
        self.assertEqual(0, interpret('M1:x0:=x0-1;M2:HALT;'))
        self.assertEqual(0, interpret('M1:x0:=x1-1;M2:HALT;'))
        self.assertEqual(0, interpret('M1:x0:=x2-6;M2:HALT;'))

    def test_assignment_number(self):
        self.assertEqual(5, interpret('M1:x0:=5;M2:HALT;'))
        self.assertEqual(2, interpret('M1:x0:=2;M2:HALT;'))
        self.assertEqual(3, interpret('M1:x0:=3;M2:HALT;'))

    def test_assignment_variable(self):
        self.assertEqual(1, interpret('M1:x0:=x0+1;M2:HALT;'), 1)
        self.assertEqual(4, interpret('M1:x0:= 5; M2:x0:=x0-1; M3:HALT;'))
        self.assertEqual(1, interpret('M1:x0:=x1-1; M2:x0:=x0+1; M3:HALT;'))
        self.assertEqual(3, interpret('M1:x0:=4; M2:x0:= x0 − 1; M3:HALT;'))

    def test_assignment_wrong_syntax(self):
        with self.assertRaises(SyntaxError):
            interpret('M1:x1:=x2;M2:HALT;')
        with self.assertRaises(SyntaxError):
            interpret('M1:x1:=0+x2;M2:HALT;')
        with self.assertRaises(SyntaxError):
            interpret('M1:x5:=-1+x4;M2:HALT;')
        with self.assertRaises(SyntaxError):
            interpret('M1:x5:=-x3+x1;M2:HALT;')
        with self.assertRaises(SyntaxError):
            interpret('M1:x5:=x1-x3;M2:HALT;')
        with self.assertRaises(SyntaxError):
            interpret('M1:x2:=x1+x4;M2:HALT;')
        with self.assertRaises(SyntaxError):
            interpret('M1:x2:=x1+2;;M2:HALT;')
        with self.assertRaises(SyntaxError):
            interpret('M1:x1:=c;M2:HALT;')
        with self.assertRaises(SyntaxError):
            interpret('M1:xi:=2;M2:HALT;')
        with self.assertRaises(SyntaxError):
            interpret('M1:x0:=xj+1;M2:HALT;')

    def test_goto_skip_lines(self):
        self.assertEqual(2, interpret('M1:x0:=2;M2:GOTO M4;M3:x0:=0;M4:HALT;'))
        self.assertEqual(3, interpret('M1:x0:=3;M2:GOTO M5;M3:GOTO M1;M4:x0:=5;M5:HALT;'))
        self.assertEqual(0, interpret('M1:GOTO M3;M2:x0:=x2+5;M3:HALT;'))

    def test_if(self):
        self.assertEqual(10, interpret('M1:x2:=10;' +
                                       'M2:IF x2=0 THEN GOTO M6;' +
                                       'M3:x0:=x0+1;M4:x2:=x2-1;' +
                                       'M5:GOTO M2; M6:HALT;'))
        self.assertEqual(1, interpret('M1:IF x0=5 THEN GOTO M4;M2:IF x0=2 THEN GOTO M4;M3:x0:=x0+1;M4:HALT;'))

    def test_syntax_missing_operator(self):
        with self.assertRaises(SyntaxError):
            interpret('M1:x0:=x1 2;M2:HALT;')

    def test_markers_start_with_one(self):
        with self.assertRaises(SyntaxError):
            interpret('M0:x0:=2;M1:HALT;')
        with self.assertRaises(SyntaxError):
            interpret('M2:x0:=2;M3:HALT;')
        with self.assertRaises(SyntaxError):
            interpret('M10:x0:=2;M11:HALT;')

    def test_markers_increment_by_one(self):
        with self.assertRaises(SyntaxError):
            interpret('M1:x0:=3;M3:HALT;')
        with self.assertRaises(SyntaxError):
            interpret('M1:x0:=2;M1:HALT;')
        with self.assertRaises(SyntaxError):
            interpret('M1: GOTO M3; M3:HALT;')

    def test_linebreaks(self):
        self.assertEqual(0, interpret('''
        M1:x1:=2;
        M2: IF x1=2 THEN GOTO M4;
        M3:HALT;
        M4:GOTO M3;'''))

    @mock.patch('loopinterpreter.input', side_effect=input_continue)
    def test_break(self, custom_input):
        self.assertEqual(2, interpret('M1:x0:=2;BREAK M2:HALT;'))

    @mock.patch('loopinterpreter.input', side_effect=input_exit)
    def test_break_exit(self, custom_input):
        self.assertEqual(-1, interpret('M1:GOTO M1; BREAK'))

    def test_assignment_missing_tokens(self):
        with self.assertRaises(SyntaxError):
            interpret('M1: x0')
        with self.assertRaises(SyntaxError):
            interpret('M1: x1 x2+2;')
        with self.assertRaises(SyntaxError):
            interpret('M1: x1 := x2+2 M2')
        with self.assertRaises(SyntaxError):
            interpret('M1: GOTO M3; M2:x1 x2+0; M3: HALT;')
        with self.assertRaises(SyntaxError):
            interpret('M1:GOTO M3; M2:x1:=2 M3: HALT;')
        with self.assertRaises(SyntaxError):
            interpret('M1:GOTO M3; M2: x1:= M3: HALT;')
        with self.assertRaises(SyntaxError):
            interpret('M1: GOTO M3; M2: x1 := x2 M3: HALT;')
        with self.assertRaises(SyntaxError):
            interpret('M1: GOTO M3; M2: x1 := x2 - M3: HALT;')
        with self.assertRaises(SyntaxError):
            interpret('M1: GOTO M3; M2: x1:= x2 + 1 M3: HALT;')

    def test_goto_missing_tokens(self):
        with self.assertRaises(SyntaxError):
            interpret('M1:GOTO :HALT;')
        with self.assertRaises(SyntaxError):
            interpret('M1: GOTO M2' +
                      'M2:HALT;')
        with self.assertRaises(SyntaxError):
            interpret('M1:GOTO M3; M2: GOTO :HALT; M3: HALT;')
        with self.assertRaises(SyntaxError):
            interpret('M1: GOTO M3;' +
                      'M2: GOTO M3' +
                      'M3:HALT;')

    def test_goto_jump_not_defined(self):
        with self.assertRaises(SyntaxError):
            interpret('M1: GOTO M2;')
        with self.assertRaises(SyntaxError):
            interpret('M1:GOTO M3; M2: GOTO M4; M3:HALT;')

    def test_halt_missing_tokens(self):
        with self.assertRaises(SyntaxError):
            interpret('M1:HALT M2:HALT;')
        with self.assertRaises(SyntaxError):
            interpret('M1:GOTO M3; M2: HALT M3:HALT;')

    def test_if_missing_tokens(self):
        with self.assertRaises(SyntaxError):
            interpret('M1: IF THEN GOTO M2; M2:HALT;')
        with self.assertRaises(SyntaxError):
            interpret('M1: IF x1 THEN GOTO M2; M2:HALT;')
        with self.assertRaises(SyntaxError):
            interpret('M1: IF x2 = THEN GOTO M2; M2:HALT;')
        with self.assertRaises(SyntaxError):
            interpret('M1: IF x0 = 0 GOTO M2; M2:HALT;')
        with self.assertRaises(SyntaxError):
            interpret('M1: IF x3 = 5 THEN M2; M2:HALT;')
        with self.assertRaises(SyntaxError):
            interpret('M1:GOTO M3; M2: IF THEN GOTO M3; M3:HALT;')
        with self.assertRaises(SyntaxError):
            interpret('M1:GOTO M3; M2: IF x1 THEN GOTO M2; M3:HALT;')
        with self.assertRaises(SyntaxError):
            interpret('M1:GOTO M3; M2: IF x2 = THEN GOTO M2; M3:HALT;')
        with self.assertRaises(SyntaxError):
            interpret('M1:GOTO M3; M2: IF x0 = 0 GOTO M2; M3:HALT;')
        with self.assertRaises(SyntaxError):
            interpret('M1:GOTO M3; M2: IF x3 = 5 THEN M2; M3:HALT;')

    def test_line_missing_tokens(self):
        with self.assertRaises(SyntaxError):
            interpret('M1 GOTO M2; M2: HALT;')
        with self.assertRaises(SyntaxError):
            interpret('M1: GOTO M3; GOTO M3; M3:HALT;')
        with self.assertRaises(SyntaxError):
            interpret('M1: GOTO M3; M2 GOTO M3; M3: HALT;')
        with self.assertRaises(SyntaxError):
            interpret('M1: GOTO M3; M2: M3; M3: HALT;')

    def test_end_without_halt(self):
        with self.assertRaises(SyntaxError):
            interpret('M1: x0:=2;')

    @mock.patch('whileinterpreter.input', side_effect=str_yes)
    def test_infinite_loop(self, custom_input):
        self.assertEqual(-1, interpret('M1: x1:=100000; M2: x1:=x1-1; M3: IF x1=0 THEN GOTO M5; M4: GOTO M1; M5:HALT;',
                                       timeout=1))

    def test_given_values(self):
        self.assertEqual(5, interpret('M1: x0:=x1+0; M2: IF x2=0 THEN GOTO M6; M3: x0:=x0+1;' +
                                      'M4: x2:=x2-1; M5: GOTO M2; M6: HALT;', [2, 3]))
        self.assertEqual(4, interpret('M1: x0:=x1+0; M2: IF x2=0 THEN GOTO M6; M3: x0:=x0+1;' +
                                      'M4: x2:=x2-1; M5: GOTO M2; M6: HALT;', [0, 4]))
        self.assertEqual(1, interpret('M1: x0:=x0+1; M2: HALT;', [2,3,4]))

    def test_wrong_given_values(self):
        with self.assertRaises(ValueError):
            interpret('M1: x0:=2;', [-1])
        with self.assertRaises(ValueError):
            interpret('M1: x0:=1;', [0.2])
        with self.assertRaises(ValueError):
            interpret('M1: x0:=4;', ['EINS'])

    def test_homeomorphic_embedding(self):
        with self.assertRaises(SyntaxError):
            interpret('M1: x1:=1; M2: IF x1 = 0 THEN GOTO M5;'
                      'M3: x1:=x1+1; M4: GOTO M2; M5: HALT;', timeout=0, debug=True)
        with self.assertRaises(SyntaxError):
            interpret('M1: x1:=100; M2: IF x1 = 0 THEN GOTO M5;'
                      'M3: x1:=x1+1; M4: GOTO M2; M5: HALT;', timeout=0, debug=True)
